from datetime import datetime
from dataclasses import dataclass
from typing import Dict, Optional
import json


@dataclass
class LocalLoggingOptions:
  path: Optional[str] = None
  time: bool = True
  prompt: bool = True
  response: bool = True
  error: bool = True


def log_generate_text(
    logging_options: LocalLoggingOptions,
    provider,
    model,
    start_time: datetime,
    end_time: datetime,
    params: Dict = None,
    prompt: Optional[str] = None,
    response: Optional[str] = None,
    error: Optional[str] = None):
  result = {
    'provider': provider,
    'model': model,
    'params': params,
  }
  if logging_options.time:
    result['start_time'] = start_time.strftime("%Y-%m-%d %H:%M:%S.%f")
    result['end_time'] = end_time.strftime("%Y-%m-%d %H:%M:%S.%f")
  if logging_options.prompt and prompt is not None:
    result['prompt'] = prompt
  if logging_options.response and response is not None:
    result['response'] = response
  if logging_options.error and error is not None:
    result['error'] = error
  with open(logging_options.path, 'a') as f:
    f.write(json.dumps(result) + '\n')
  f.close()



import os
from typing import Dict, Optional
import proxai.types as types
from proxai.connectors.model_connector import ModelConnector
from proxai.connectors.openai import OpenAIConnector
from proxai.connectors.claude import ClaudeConnector
from proxai.connectors.gemini import GeminiConnector
from proxai.connectors.cohere_api import CohereConnector
from proxai.connectors.databricks import DatabricksConnector
from proxai.connectors.mistral import MistralConnector
from proxai.connectors.hugging_face import HuggingFaceConnector
from proxai.logging.utils import LocalLoggingOptions

_RUN_TYPE: types.RunType = types.RunType.PRODUCTION
_REGISTERED_SIGNATURE: types.ModelSignature = None
_MODEL_CONNECTOR: ModelConnector = None
_LOCAL_LOGGING_OPTIONS: LocalLoggingOptions = LocalLoggingOptions()


def _set_run_type(run_type: types.RunType):
  global _RUN_TYPE
  _RUN_TYPE = run_type


def _get_model_connector_from_signature(
    signature: types.ModelSignature) -> ModelConnector:
  global _LOCAL_LOGGING_OPTIONS
  connector = None
  if signature.provider == types.Provider.OPENAI:
    connector =  OpenAIConnector
  elif signature.provider == types.Provider.CLAUDE:
    connector =  ClaudeConnector
  elif signature.provider == types.Provider.GEMINI:
    connector =  GeminiConnector
  elif signature.provider == types.Provider.COHERE:
    connector =  CohereConnector
  elif signature.provider == types.Provider.DATABRICKS:
    connector =  DatabricksConnector
  elif signature.provider == types.Provider.MISTRAL:
    connector =  MistralConnector
  elif signature.provider == types.Provider.HUGGING_FACE:
    connector =  HuggingFaceConnector
  else:
    raise ValueError("Provider not supported")

  if _LOCAL_LOGGING_OPTIONS.path:
    return connector(
        signature=signature,
        run_type=_RUN_TYPE,
        logging_options=_LOCAL_LOGGING_OPTIONS)

  return connector(
      signature=signature,
      run_type=_RUN_TYPE)

def _get_registered_model_connector() -> ModelConnector:
  global _REGISTERED_SIGNATURE
  global _MODEL_CONNECTOR
  if not _REGISTERED_SIGNATURE:
    _REGISTERED_SIGNATURE = types.ModelSignature(
        provider=types.Provider.OPENAI,
        model=types.OpenAIModel.GPT_3_5_TURBO)
    _MODEL_CONNECTOR = _get_model_connector_from_signature(
        _REGISTERED_SIGNATURE)
  return _MODEL_CONNECTOR


def local_logging_path(
      path: str,
      time: bool = True,
      prompt: bool = True,
      response: bool = True,
      error: bool = True):
  global _LOCAL_LOGGING_OPTIONS
  _LOCAL_LOGGING_OPTIONS.path = path
  _LOCAL_LOGGING_OPTIONS.time = time
  _LOCAL_LOGGING_OPTIONS.prompt = prompt
  _LOCAL_LOGGING_OPTIONS.response = response
  _LOCAL_LOGGING_OPTIONS.error = error


def register_model(provider: str, model: str):
  global _REGISTERED_SIGNATURE
  global _MODEL_CONNECTOR

  providers = set(item.value for item in types.Provider)
  if provider not in providers:
    raise ValueError(
      f'Provider not supported: {provider}. Supported providers: {providers}')

  if provider in [types.Provider.OPENAI,
                  types.Provider.CLAUDE,
                  types.Provider.GEMINI,
                  types.Provider.COHERE,
                  types.Provider.DATABRICKS,
                  types.Provider.MISTRAL,
                  types.Provider.HUGGING_FACE]:
    models = set(item.value for item in types.MODEL_MAP[provider])
    if model not in models:
      raise ValueError(
        f'Model {model} not supported for provider {provider}.\n'
        f'Supported models: {models}')
    _REGISTERED_SIGNATURE = types.ModelSignature(
        provider=types.Provider(provider),
        model=types.MODEL_MAP[provider](model))
  else:
    raise ValueError(
      'Provider not supported yet. We are looking for contributors!\n'
      f'Provider: {provider}')

  _MODEL_CONNECTOR = _get_model_connector_from_signature(
      _REGISTERED_SIGNATURE)


def generate_text(
    prompt: str,
    max_tokens: int = 100) -> str:
  model_connector = _get_registered_model_connector()
  return model_connector.generate_text(prompt, max_tokens)



from typing import Dict, List


class _MockResponse(object):
  message: str

  def __init__(self):
    self.message = 'mock response'


class DatabricksMock(object):
  def create(self, model: str, messages: List[Dict]) -> _MockResponse:
    return _MockResponse()



from typing import Any, Dict, List


class _MockMessage(object):
  content: str


class _MockChoice(object):
  message: _MockMessage

  def __init__(self):
    self.message = _MockMessage()
    self.message.content = 'mock response'


class _MockResponse(object):
  choices: List[_MockChoice]

  def __init__(self):
    self.choices = [_MockChoice()]


class MistralMock(object):
  def chat(self, model: str, messages: List[Any]) -> _MockResponse:
    return _MockResponse()



from typing import Union, Optional
import cohere
import proxai.types as types
from .cohere_api_mock import CohereMock
from .model_connector import ModelConnector


class CohereConnector(ModelConnector):
  def init_model(self):
    return cohere.Client()

  def init_mock_model(self):
    return CohereMock()

  def generate_text_proc(self, prompt: str, max_tokens: int) -> str:
    response = self.api.chat(
        message=prompt,
        model=self.model_signature.model.value)
    return response.text



import anthropic
from typing import Union, Optional
import proxai.types as types
from .claude_mock import ClaudeMock
from .model_connector import ModelConnector


class ClaudeConnector(ModelConnector):
  def init_model(self):
    return anthropic.Anthropic()

  def init_mock_model(self):
    return ClaudeMock()

  def generate_text_proc(self, prompt: str, max_tokens: int) -> str:
    message = self.api.messages.create(
        model=self.model_signature.model.value,
        max_tokens=1024,
        messages=[
            {'role': 'user', 'content': prompt}
        ]
    )
    return message.content[0].text



from typing import Dict, List


class _MockResponse(object):
  text: str

  def __init__(self):
    self.text = 'mock response'


class CohereMock(object):
  def chat(self, message: str, model: str) -> _MockResponse:
    return _MockResponse()



from datetime import datetime
from typing import Any, Dict, Optional
from proxai.types import ModelSignature, RunType
from proxai.logging.utils import log_generate_text


class ModelConnector(object):
  model_signature: Optional[ModelSignature] = None
  run_type: RunType
  _api: Optional[Any] = None
  _logging_options: Optional[Dict] = None

  def __init__(
      self,
      signature: ModelSignature,
      run_type: RunType,
      logging_options: Optional[dict] = None):
    self.model_signature = signature
    self.run_type = run_type
    if logging_options:
      self._logging_options = logging_options

  @property
  def api(self):
    if not self._api:
      if self.run_type == RunType.PRODUCTION:
        self._api = self.init_model()
      else:
        self._api = self.init_mock_model()
    return self._api

  def init_model(self):
    raise NotImplementedError

  def init_mock_model(self):
    raise NotImplementedError

  def _log_generate_text(
      self,
      start_time: datetime,
      end_time: datetime,
      prompt: Optional[str] = None,
      response: Optional[str] = None,
      error: Optional[str] = None):
    if self._logging_options:
      log_generate_text(
          logging_options=self._logging_options,
          provider=self.model_signature.provider,
          model=self.model_signature.model,
          start_time=start_time,
          end_time=end_time,
          prompt=prompt,
          response=response,
          error=error)

  def generate_text(self, prompt: str, max_tokens: int) -> str:
    start_time = datetime.now()
    try:
      response =  self.generate_text_proc(prompt, max_tokens)
    except Exception as e:
      self._log_generate_text(
          start_time=start_time,
          end_time=datetime.now(),
          prompt=prompt,
          error=str(e))
      raise e

    self._log_generate_text(
        start_time=start_time,
        end_time=datetime.now(),
        prompt=prompt,
        response=response)
    return response

  def generate_text_proc(self, prompt: str, max_tokens: int) -> dict:
    raise NotImplementedError



from typing import Union, Optional
from openai import OpenAI
import proxai.types as types
from .openai_mock import OpenAIMock
from .model_connector import ModelConnector


class OpenAIConnector(ModelConnector):
  def init_model(self):
    return OpenAI()

  def init_mock_model(self):
    return OpenAIMock()

  def generate_text_proc(self, prompt: str, max_tokens: int) -> str:
    completion = self.api.chat.completions.create(
        model=self.model_signature.model.value,
        messages=[
          {'role': 'system', 'content': 'You are an helpful assistant.'},
          {'role': 'user', 'content': prompt}
        ]
    )
    return completion.choices[0].message.content



from typing import Dict, List


class _MockMessage(object):
  content: str


class _MockChoice(object):
  message: _MockMessage

  def __init__(self):
    self.message = _MockMessage()
    self.message.content = 'mock response'


class _MockResponse(object):
  choices: List[_MockChoice]

  def __init__(self):
    self.choices = [_MockChoice()]


class _MockCompletions(object):
  def create(self, model: str, messages: List[Dict]) -> _MockResponse:
    return _MockResponse()


class _MockChat(object):
  completions: _MockCompletions

  def __init__(self):
    self.completions = _MockCompletions()


class OpenAIMock(object):
  chat: _MockChat

  def __init__(self):
    self.chat = _MockChat()



from typing import Union, Optional
import google.generativeai as genai
import proxai.types as types
from .gemini_mock import GeminiMock
from .model_connector import ModelConnector


class GeminiConnector(ModelConnector):
  def init_model(self):
    return genai.GenerativeModel(self.model_signature.model.value)

  def init_mock_model(self):
    return GeminiMock()

  def generate_text_proc(self, prompt: str, max_tokens: int) -> str:
    response = self.api.generate_content(prompt)
    return response.text



import os
import requests
from typing import Union, Optional
import proxai.types as types
from .hugging_face_mock import HuggingFaceMock
from .model_connector import ModelConnector


class _HuggingFaceRequest:
  def __init__(self):
    self.api_url = 'https://api-inference.huggingface.co/models/'
    self.headers = {
        'Authorization': f'Bearer {os.environ["HUGGINGFACE_API_KEY"]}'}

  def generate_content(self, prompt: str, model: str) -> str:
    response = requests.post(
        self.api_url + model,
        headers=self.headers,
        json={'inputs': prompt})
    text = response.json()[0]['generated_text']
    if text.startswith(prompt):
      text = text[len(prompt):]
    return text


class HuggingFaceConnector(ModelConnector):
  def init_model(self):
    return _HuggingFaceRequest()

  def init_mock_model(self):
    return HuggingFaceMock()

  def generate_text_proc(self, prompt: str, max_tokens: int) -> str:
    response = self.api.generate_content(
        prompt, model=self.model_signature.model.value)
    return response



from typing import Union, Optional
from mistralai.client import MistralClient
from mistralai.models.chat_completion import ChatMessage
import proxai.types as types
from .mistral_mock import MistralMock
from .model_connector import ModelConnector


class MistralConnector(ModelConnector):
  def init_model(self):
    return MistralClient()

  def init_mock_model(self):
    return MistralMock()

  def generate_text(self, prompt: str, max_tokens: int) -> str:
    response = self.api.chat(
        model=self.model_signature.model.value,
        messages=[
            ChatMessage(role='user', content=prompt)
        ],
    )
    return response.choices[0].message.content



import dataclasses
import enum
from typing import Union


class RunType(enum.Enum):
  PRODUCTION = 1
  TEST = 2


class Provider(str, enum.Enum):
  OPENAI = 'openai'
  CLAUDE = 'claude'
  GEMINI = 'gemini'
  COHERE = 'cohere'
  DATABRICKS = 'databricks'
  MISTRAL = 'mistral'
  HUGGING_FACE = 'hugging_face'


class OpenAIModel(str, enum.Enum):
  """OpenAI models.

  Models provided by OpenAI:
  https://platform.openai.com/docs/guides/text-generation
  """
  # Newer models (2023–)
  GPT_4 = 'gpt-4'
  GPT_4_TURBO_PREVIEW = 'gpt-4-turbo-preview'
  GPT_3_5_TURBO = 'gpt-3.5-turbo'

  # Updated legacy models (2023)
  GPT_3_5_TURBO_INSTRUCT = 'gpt-3.5-turbo-instruct'
  BABBAGE_002 = 'babbage-002'
  DAVINCI_002 = 'davinci-002'


class ClaudeModel(str, enum.Enum):
  """Claude models.

  Models provided by Claude:
  https://claude.ai/docs/models
  """
  # Latest models (03/28/2024)
  CLAUDE_3_OPUS =  'claude-3-opus-20240229'
  CLAUDE_3_SONNET = 'claude-3-sonnet-20240229'
  CLAUDE_3_HAIKU = 'claude-3-haiku-20240307'


class GeminiModel(str, enum.Enum):
  """Gemini models.

  Models provided by Gemini:
  https://ai.google.dev/models/gemini
  """
  GEMINI_1_0_PRO = 'models/gemini-1.0-pro'
  GEMINI_1_0_PRO_001 =  'models/gemini-1.0-pro-001'
  GEMINI_1_0_PRO_LATEST = 'models/gemini-1.0-pro-latest'
  GEMINI_1_0_PRO_VISION_LATEST = 'models/gemini-1.0-pro-vision-latest'
  GEMINI_PRO = 'models/gemini-pro'
  GEMINI_PRO_VISION = 'models/gemini-pro-vision'


class CohereModel(str, enum.Enum):
  """Cohere models.

  Models provided by Cohere:
  https://docs.cohere.com/docs/models
  """
  COMMAND_LIGHT = 'command-light'
  COMMAND_LIGHT_NIGHTLY = 'command-light-nightly'
  COMMAND = 'command'
  COMMAND_NIGHTLY = 'command-nightly'
  COMMAND_R = 'command-r'


class DatabricksModel(str, enum.Enum):
  """Databricks models.

  Models provided by Databricks:
  https://docs.databricks.com/en/machine-learning/foundation-models/index.html#provisioned-throughput-foundation-model-apis
  """
  DATABRICKS_DBRX_INSTRUCT = 'databricks-dbrx-instruct'
  DATABRICKS_MIXTRAL_8x7b_INSTRUCT = 'databricks-mixtral-8x7b-instruct'
  DATABRICKS_LLAMA_2_70b_CHAT = 'databricks-llama-2-70b-chat'


class MistralModel(str, enum.Enum):
  """Mistral models.

  Models provided by Mistral:
  https://docs.mistral.ai/platform/endpoints/
  """
  OPEN_MISTRAL_7B = 'open-mistral-7b'
  OPEN_MIXTRAL_8X7B = 'open-mixtral-8x7b'
  MISTRAL_SMALL_LATEST = 'mistral-small-latest'
  MISTRAL_MEDIUM_LATEST = 'mistral-medium-latest'
  MISTRAL_LARGE_LATEST = 'mistral-large-latest'


class HuggingFaceModel(str, enum.Enum):
  """Hugging Face models.

  Models provided by Hugging Face on HuggingFaceChat:
  https://huggingface.co/chat/models
  """
  # To be able to use Google models, you need to sign terms of service:
  GOOGLE_GEMMA_7B_IT = 'google/gemma-7b-it'
  # Requires pro subscription:
  # META_LLAMA_2_70B_CHAT_HF = 'meta-llama/Llama-2-70b-chat-hf'
  # Requires pro subscription:
  # CODELLAMA_70B_INSTRUCT_HF = 'codellama/CodeLlama-70b-Instruct-hf'
  MISTRAL_MIXTRAL_8X7B_INSTRUCT_V01 = 'mistralai/Mixtral-8x7B-Instruct-v0.1'
  MISTRAL_MISTRAL_7B_INSTRUCT_V02 = 'mistralai/Mistral-7B-Instruct-v0.2'
  NOUS_HERMES_2_MIXTRAL_8X7B_DPO = 'NousResearch/Nous-Hermes-2-Mixtral-8x7B-DPO'
  OPENCHAT_3_5 = 'openchat/openchat-3.5-0106'


MODEL_MAP = {
    Provider.OPENAI: OpenAIModel,
    Provider.CLAUDE: ClaudeModel,
    Provider.GEMINI: GeminiModel,
    Provider.COHERE: CohereModel,
    Provider.DATABRICKS: DatabricksModel,
    Provider.MISTRAL: MistralModel,
    Provider.HUGGING_FACE: HuggingFaceModel,
}


GENERATE_TEXT_MODELS = {
    Provider.OPENAI: [
        OpenAIModel.GPT_4,
        OpenAIModel.GPT_4_TURBO_PREVIEW,
        OpenAIModel.GPT_3_5_TURBO,
    ],
    Provider.CLAUDE: [
        ClaudeModel.CLAUDE_3_OPUS,
        ClaudeModel.CLAUDE_3_SONNET,
        ClaudeModel.CLAUDE_3_HAIKU,
    ],
    Provider.GEMINI: [
        GeminiModel.GEMINI_1_0_PRO,
        GeminiModel.GEMINI_1_0_PRO_001,
        GeminiModel.GEMINI_1_0_PRO_LATEST,
        GeminiModel.GEMINI_PRO,
    ],
    Provider.COHERE: [
        CohereModel.COMMAND_LIGHT,
        CohereModel.COMMAND_LIGHT_NIGHTLY,
        CohereModel.COMMAND,
        CohereModel.COMMAND_NIGHTLY,
        CohereModel.COMMAND_R,
    ],
    Provider.DATABRICKS: [
        DatabricksModel.DATABRICKS_DBRX_INSTRUCT,
        DatabricksModel.DATABRICKS_MIXTRAL_8x7b_INSTRUCT,
        DatabricksModel.DATABRICKS_LLAMA_2_70b_CHAT,
    ],
    Provider.MISTRAL: [
        MistralModel.OPEN_MISTRAL_7B,
        MistralModel.OPEN_MIXTRAL_8X7B,
        MistralModel.MISTRAL_SMALL_LATEST,
        MistralModel.MISTRAL_MEDIUM_LATEST,
        MistralModel.MISTRAL_LARGE_LATEST,
    ],
    Provider.HUGGING_FACE: [
        HuggingFaceModel.GOOGLE_GEMMA_7B_IT,
        HuggingFaceModel.MISTRAL_MIXTRAL_8X7B_INSTRUCT_V01,
        HuggingFaceModel.MISTRAL_MISTRAL_7B_INSTRUCT_V02,
        HuggingFaceModel.NOUS_HERMES_2_MIXTRAL_8X7B_DPO,
        HuggingFaceModel.OPENCHAT_3_5,
    ],
}


@dataclasses.dataclass
class ModelSignature:
  provider: Provider
  model: Union[
      OpenAIModel,
      ClaudeModel,
      GeminiModel,
      CohereModel]



from typing import Dict, List
# [ContentBlock(text="Hello! It's nice to meet you. How can I assist you today?", type='text')]

class _MockContentBlock(object):
  text: str
  type: str

  def __init__(self):
    self.text = 'mock response'
    self.type = 'text'


class _MockResponse(object):
  content: List[_MockContentBlock]

  def __init__(self):
    self.content = [_MockContentBlock()]


class _MockMessages(object):
  def create(
      self,
      model: str,
      max_tokens: int,
      messages: List[Dict]) -> _MockResponse:
    return _MockResponse()


class ClaudeMock(object):
  messages: _MockMessages

  def __init__(self):
    self.messages = _MockMessages()



from typing import Union, Optional, Type
from databricks_genai_inference import ChatCompletion
import proxai.types as types
from .databricks_mock import DatabricksMock
from .model_connector import ModelConnector


class DatabricksConnector(ModelConnector):
  def init_model(self):
    return ChatCompletion

  def init_mock_model(self):
    return DatabricksMock()

  def generate_text_proc(self, prompt: str, max_tokens: int) -> str:
    response = self.api.create(
        model=self.model_signature.model.value,
        messages=[
            {'role': 'system', 'content': 'You are a helpful assistant.'},
            {'role': 'user', 'content': prompt}])
    return response.message
